from Point import Point
from Surface import Surface
from UnitSurfacePoints import UnitSurfacePoints

# Using 'private' calls such as atom._x instead of atom.x for efficiency
# (bypasses the get_x function)

AtomError = "atom error"

class Atom(Point):
    """Class Atom

    The Atom class is one of the largest classes. It inherits all the
    attributes and methods of the Point class. It also holds a reference to an
    intance of the Surface class, the surface attribute.

    The MacroMoleculeAtom, LigandAtom, WaterProbe and VoidProbe all inherit
    the attributes and methods of the Atom class.

    The attribute atom_org_id does not make sense for the VoidProbes and the
    WaterProbes added using the Assembly.addWater() method and they will
    therefore also be left as None.  Finally the attribute atom_polarity does
    not make sense for the VoidProbes and will be left as None. The attribute
    atom_org_name and atom_type will both be set to 'Probe' for the
    VoidProbes.

    The neighbor list attibutes are listes of references to neighboring atoms.
    The far_neighbor attribute is used by the set_accessibility() method.

    The various radius attributes (far_radius, close_radius, probe_radius,
    char_apol_radius, char_pol_radius, char_pol_radius_max) are all set by the
    MacroMoleculeAtom/LigandAtom/WaterProbe functions setRadius(). The
    VoidProbe sets the _probe_radius and _char_apol_radius during the
    Assembly.fillProbe() method.

    The _atom_polarity is set by the MacroMoleculeAtom/LigandAtom/WaterProbe
    functions set_polarity(). In order to be consistent with the integers used
    to define the accessibility, the atom_polarity integers are defined as:
    1) Polar  -- (0)
    2) Apolar -- (2)

    The include attribute is set to 1 by default when the atoms are read in
    using the MacroMoleculeAtom/LigandAtom/WaterProbe/VoidProbe atomInput()
    method. In cases where the user specifies use onlyBinindingSite or use
    onlyInterface with the methods Assembly.onlyBinindingSite() and
    Assembly.onlyInterface() respectively any atoms not part of the binding
    site or interface will have their inlude attribute set to 0.

    The sub_atom_type attribute tells the atom to which type of sub atom it
    belongs. The sub atom types are defined as:
    1) MacroMoleculeAtom -- (1)
    2) Ligand            -- (2)
    3) WaterProbe        -- (3)
    4) VoidProbe         -- (4)

    The molecule insance tells the atom if it is part of file1 or file2.

    Attributes:
    atom_org_id          -- original id of atom
    atom_org_name        -- original name of atom
    atom_type            -- atom type 
    atom_polarity        -- atom polarity (polar (0) or apolar (2))
    char_apol_radius     -- characteristic apolar radius of atom
    char_pol_radius      -- characteristic polar radius of atom
    char_pol_radius_max  -- maximum charchteristic polar radius of atom
    close_neighbors      -- list of close neighbors
    far_neighbors        -- list of far neighbors
    close_radius         -- 
    far_radius           --
    probe_radius         --
    include              -- should the atom be used in calculation or not
    surface              -- instance of a class Surface
    sub_atom_type        -- what type of molecule does the atom belong to
    molecule_instance    -- which file does the atom belong to

    Methods:
    set_surface_points    -- initiates an instance of the class Surface to
                             attribute the surface
    set_accessibility     -- set the accessibility of the surface points
    append_close_neighbor -- appends an atom to the close neighbor list
    append_far_neighbor   -- appends an atom to the far neighbor list

    Examples:
    >>> from Atom import Atom
    >>> from UnitSurfacePoints import UnitSurfacePoints
    >>> u = UnitSurfacePoints()
    >>>
    >>> a1 = Atom(1, ' O  ', 'O', 14.345, -5.635, 26.750, 1, 0)
    >>> a1.far_radius = 1.62 + 1.97
    >>> a1.close_radius = 1.35 + 1.50
    >>> a1.probe_radius = 2.06
    >>> a1.atom_polarity = 0
    >>> a1.char_apol_radius = 2.06
    >>> a1.char_pol_radius = 1.35
    >>> a1.char_pol_radius_max = 1.62
    >>>
    >>> a2 = Atom(2, ' C  ', 'C', 12.891, -4.205, 27.064, 1, 0)
    >>> a2.far_radius = 1.93 + 1.97
    >>> a2.close_radius = 1.63 + 1.50
    >>> a2.probe_radius = 2.23
    >>> a2.atom_polarity = 2
    >>> a2.char_apol_radius = 1.93
    >>> a2.set_surface_points(u, a1.close_radius, 0, 0)
    >>>
    >>> a1.append_close_neighbor(a2)
    >>> a1.append_far_neighbor(a2)
    >>> a2.append_close_neighbor(a1)
    >>> a2.append_far_neighbor(a1)
    >>>
    >>> a1.set_surface_points(u, a1.close_radius, 0, 0)
    >>> a2.set_surface_points(u, a2.close_radius, 0, 0)
    >>>
    >>> a1.set_accessibility(u)
    >>> a2.set_accessibility(u)
    >>>
    >>> a1.in_contact(a2)
    2

    """
    
    def __init__(self, atom_org_id, atom_org_name, atom_type, x, y, z,
                 sub_atom_type, molecule_instance):
        """
        Initialise the object.

        atom_org_id: int
        atom_org_name: string
        atom_type: string
        x: float
        y: float
        z: float
        sub_atom_type: int
        molecule_instance: int

        This method sets the following attributes of the atom:
        1) _atom_org_id
        2) _atom_org_name
        3) _atom_type
        4) _x, _y, _z
        5) _sub_atom_type
        6) _molecule_instance

        """
        self._close_neighbors = []
        self._far_neighbors = []

        self._atom_org_id = atom_org_id
        self._atom_org_name = atom_org_name
        self._atom_type = atom_type
        self._sub_atom_type = sub_atom_type
        self._molecule_instance = molecule_instance
        Point.__init__(self, x, y, z)
    
    @property
    def atom_org_id(self):
        """Get atom_org_id attribute."""
        return self._atom_org_id
    
    @property
    def atom_org_name(self):
        """Get atom_org_name attribute."""
        return self._atom_org_name
    
    @property
    def atom_type(self):
        """Get atom_type attribute."""
        return self._atom_type
    
    @property
    def surface(self):
        """Get surface attribute."""
        return self._surface
    
    @property
    def sub_atom_type(self):
        """Get sub_atom_type attribute."""
        return self._sub_atom_type
    
    @property
    def molecule_instance(self):
        """Get molecule_instance attribute."""
        return self._molecule_instance
    
    @property
    def far_neighbors(self):
        """Get far_neighbors attribute."""
        return self._far_neighbors

    @property
    def close_neighbors(self):
        """Get close_neighbors attribute."""
        return self._close_neighbors

    def get_atom_polarity(self):
        """Get atom_polarity attribute."""
        return self._atom_polarity
    
    def set_atom_polarity(self, atom_polarity):
        """Set atom_polarity attribute."""
        self._atom_polarity = int(atom_polarity)
    
    atom_polarity = property(get_atom_polarity, set_atom_polarity)
    
    def get_char_apol_radius(self):
        """Get char_apol_radius attribute."""
        return self._char_apol_radius
    
    def set_char_apol_radius(self, char_apol_radius):
        """Set char_apol_radius attribute."""
        self._char_apol_radius = float(char_apol_radius)
    
    char_apol_radius = property(get_char_apol_radius, set_char_apol_radius)
    
    def get_char_pol_radius(self):
        """Get char_pol_radius attribute."""
        return self._char_pol_radius
    
    def set_char_pol_radius(self, char_pol_radius):
        """Set char_pol_radius attribute."""
        self._char_pol_radius = float(char_pol_radius)
    
    char_pol_radius = property(get_char_pol_radius, set_char_pol_radius)
    
    def get_char_pol_radius_max(self):
        """Get char_pol_radius_max attribute."""
        return self._char_pol_radius_max
    
    def set_char_pol_radius_max(self, char_pol_radius_max):
        """Set char_pol_radius_max attribute."""
        self._char_pol_radius_max = float(char_pol_radius_max)
    
    char_pol_radius_max = property(get_char_pol_radius_max, set_char_pol_radius_max)
    
    def get_close_radius(self):
        """Get close_radius attribute."""
        return self._close_radius
    
    def set_close_radius(self, close_radius):
        """Set close_radius attribute."""
        self._close_radius = float(close_radius)
    
    close_radius = property(get_close_radius, set_close_radius)
    
    def get_far_radius(self):
        """Get far_radius attribute."""
        return self._far_radius
    
    def set_far_radius(self, far_radius):
        """Set far_radius attribute."""
        self._far_radius = float(far_radius)
    
    far_radius = property(get_far_radius, set_far_radius)
    
    def get_probe_radius(self):
        """Get probe_radius attribute."""
        return self._probe_radius
    
    def set_probe_radius(self, probe_radius):
        """Set probe_radius attribute."""
        self._probe_radius = float(probe_radius)
    
    probe_radius = property(get_probe_radius, set_probe_radius)
    
    def get_include(self):
        """Get include attribute."""
        return self._include
    
    def set_include(self, include):
        """Set include attribute."""
        self._include = int(include)
    
    include = property(get_include, set_include)
    
    def set_surface_points(self, unit_points, radius, accessibility, flag):
        """
        This method initiates an instance of the object Surface and links it
        to the surface attribute.

        unit_points: UnitSurfacePoints
        accessibility: int
        flag: int

        The unit_points argument should be a refernce to an instance of a
        UnitSurfacePoint object.

        The radius is the disance from the Atoms centre the surface points
        should be placed.

        The accessibility should be an integer (e.g 0 for polar).

        The flag should be an integer. The flag definitions are:
        1) enabled  -- (0)
        2) disabled -- (1)
        
        """
        self._surface = Surface(unit_points, self, radius, accessibility, flag)

    def set_accessibility(self, unit_points):
        """
        This methods sets the accessibility of surface points based on
        atom_polarity and clashes to neighboring atoms.

        The accessibility definitions are:
        1) Polar solvent accessible  -- (0)
        2) Apolar solvent accessible -- (2)
        3) Probe accessible          -- (-1)
        4) Probe inaccessible        -- (1)
        
        """
        # check that neighbors, radius and polarity have been set
        if len(self._far_neighbors) < 1:  # zero neighbors = not initialised
            raise AtomError, "set_accessibility() requires far_neighbors"
        if self._close_radius == None:
            raise AtomError, "set_accessibility() requires radius to be set"
        if self._atom_polarity == None:
            raise AtomError, "set_accessibility() requires polarity to be set"

        # Draw dot surface as radius of water
        # (all surface points initially treated as if the atom was polar)
        self.set_surface_points(unit_points, self._close_radius, 0, 0)

        # Check for polar clashes with neighbors
        if self._atom_polarity == 0:
            for surface_point in self._surface._surface_point_list:
                for far_neighbor in self._far_neighbors:
                    if surface_point.clashes(far_neighbor,
                        far_neighbor._close_radius):     # probe accessible
                            surface_point.update_surface_point(
                                self._probe_radius, self._x, self._y, self._z,  -1, 0)
                            break   # out of the far_neighbor loop
        # Modify results for apolar atoms
        if self._atom_polarity == 2:
            for surface_point in self._surface._surface_point_list:

                # Check if the surface point should be treated as polar or
                # apolar
                for far_neighbor in self._far_neighbors:
                    if far_neighbor._atom_polarity == 0:  # neighbor is polar
                        if surface_point.clashes(far_neighbor,
                            far_neighbor._far_radius):  # clash
                            surface_point.flag = 1    # don't use later on
                            break   # out of far_neighbor loop

                # Update the point if it is to be treated as apolar
                if surface_point._flag == 0:        # however do use these
                    surface_point.update_surface_point(
                        self._far_radius, self._x, self._y, self._z,  2, 0)

                # check for clashes again
                for far_neighbor in self._far_neighbors:
                    cutoff = far_neighbor._far_radius     # default to apolar
                    if far_neighbor._atom_polarity == 0:  # polar
                        cutoff = far_neighbor._close_radius
                    if surface_point.clashes(far_neighbor, cutoff):
                        # probe accessible
                        surface_point.update_surface_point(
                            self._probe_radius, self._x, self._y, self._z, -1, 0)
                        break   # out of far_neighbor loop

        # Check to see if probe may be inserted on each unsolvated dot
        for surface_point in self._surface._surface_point_list:
            if surface_point._accessibility == -1:  # probe accessible
                for far_neighbor in self._far_neighbors:
                    if surface_point.clashes(far_neighbor,
                        far_neighbor._probe_radius):    # clashes
                        surface_point.set_accessibility(1)
                        surface_point.update_surface_point(                   # probe
                            self._probe_radius, self._x, self._y, self._z, 1, 0) # inaccessible
                        break   # out of far neighbor loop
        self.surface.remove_inaccessible_points()

    def append_close_neighbor(self, atom):
        """
        This method appends an atom reference to the close neighbor list.
        """
        self._close_neighbors.append(atom)
    
    def append_far_neighbor(self, atom):
        """
        This method appends an atom reference to the far neighbor list.
        """
        self._far_neighbors.append(atom)

    def insert_far_neighbor(self, atom):
        """
        This method inserts an atom reference to the far neighbor list.
        """
        self._far_neighbors.insert(0, atom)

    def clear_neighbors(self):
        """
        This method empties the far and close neighbor lists.
        """
        for atom in self._close_neighbors:
            atom._close_neighbors.remove(self)
        for atom in self._far_neighbors:
            atom._far_neighbors.remove(self)
        del self._close_neighbors[:]
        del self._far_neighbors[:]

    def clear_surface(self):
        """
        This method empties the surface attribute.
        """
        try:
            del self._surface._surface_point_list[:]
            del self._surface
        except AttributeError:
            pass

    def in_contact(self, target):
        """
        This method checks to see if the Atom is in contact with the target
        Atom. The function returns an integer based on the type of contact.
        The types of contacts are defined as:
        1) No contact     - (-1)
        2) Polar contact  - (1)
        3) Apolar contact - (2)

        target: Atom

        returns int
        """
        # Check that it is not the same atom
        if self == target:
            return -1
        # Check for distance cutoff
        separation2 = self.get_distance2(target)
        apolarCutoff = ( ( self._probe_radius
                          + target._char_apol_radius
                          + self._char_apol_radius
                          + target._probe_radius ) / 2 )
        apolarCutoff2 = ( apolarCutoff * apolarCutoff )
        
        # Final test for covalent bonding
        if separation2 <= ( target._char_apol_radius
            * target._char_apol_radius ):
            return -1

        if separation2 <= apolarCutoff2:
            # Atoms in contact

            # Test polarity
            polarity = 2     # defualt to apolar

            if target._sub_atom_type == 1 or target._sub_atom_type == 2:
                if ( self._atom_polarity == 0
                    and target._atom_polarity == 0 ):
                    # Potential polar contact
                    polarCutoff = ( self.char_pol_radius_max
                        + target.char_pol_radius_max )
                    polarCutoff2 = polarCutoff * polarCutoff
                    if separation2 <= polarCutoff2:
                        polarity = 0

            if target._sub_atom_type == 3:
                # target is a water atom
                if self._atom_polarity == 0:
                    # Potential polar contact
                    polarCutoff = ( self.char_pol_radius_max
                        + target.char_pol_radius_max )
                    polarCutoff2 = polarCutoff * polarCutoff
                    if separation2 <= polarCutoff2:
                        polarity = 0

            if target._sub_atom_type == 4:
                # target is a probe atom
                if self._atom_polarity == 0:
                    polarity = 0

            return polarity
        else:
            return -1
        
# Test by running the command 'python Atom.py' in the terminal
if __name__ == "__main__":
    import doctest
    doctest.testmod()
    import unittest
    suite = doctest.DocFileSuite('doc_tests/test_Atom.txt',
        optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE)
    unittest.TextTestRunner().run(suite)
