from Atom import Atom
# Deal with the radii
import Radii
R = Radii.Radii()

MacroMoleculeAtomError = "MacroMoleculeAtom error"

class MacroMoleculeAtom(Atom):
    """Class MacroMoleculeAtom
    
    The MacroMoleculeAtom class inherits all the attributes and methods of the
    Atom class. Furthermore it has the following extra attributes:
    1) res_original_id -- the residue's original id
    2) res_type       -- the residue's type (for example ALA, GLY etc)
    3) chain         -- the chain that the residue belongs to

    Examples:
    >>> from MacroMoleculeAtom import MacroMoleculeAtom
    >>> import Radii
    >>> R = Radii.Radii()
    >>> R.radii_set = 'LI_AND_NUSSINOV'
    >>> mma = MacroMoleculeAtom(1, '  O ', 'O', 0.0, 0.0, 0.0, 1, 0, 1, 'ALA', 'A')
    >>> mma.atom_type
    'O'
    >>> mma.res_original_id
    1
    >>> mma.res_type
    'ALA'
    >>> mma.chain
    'A'
    >>> mma.atom_polarity
    0
    >>> '%.2f' % mma.char_apol_radius
    '1.51'
    >>> '%.2f' % mma.char_pol_radius
    '1.38'
    >>> '%.2f' % mma.char_pol_radius_max
    '1.50'
    >>> '%.2f' % mma.close_radius
    '2.75'
    >>> '%.2f' % mma.far_radius
    '3.05'
    >>> '%.2f' % mma.probe_radius
    '1.77'

    """

    def __init__(self, atom_org_id, atom_org_name, atom_type, x, y, z,
                 sub_atom_type, molecule_instance, res_original_id, res_type,
                 chain):
        """
        Initialise the object.

        atom_org_id: int
        atom_org_name: string
        atom_type: string
        x: float
        y: float
        z: float
        sub_atom_type: int
        molecule_instance: int
        res_original_id: int
        res_type: string
        chain: string

        This method sets the following attributes of the atom:
        1) _atom_org_id
        2) _atom_org_name
        3) _atom_type
        4) _x, _y, _z
        5) _sub_atom_type
        6) _molecule_instance
        7) _res_original_id
        8) _res_type
        9) _chain
        10) _atom_polarity
        11) various radii...

        """
        self._res_original_id = res_original_id
        self._res_type = res_type
        self._chain = chain
        Atom.__init__(self, atom_org_id, atom_org_name, atom_type, x, y, z,
                      sub_atom_type, molecule_instance)
        self._set_atom_polarity()
        self._set_radius()
        
    @property
    def res_original_id(self):
        """Get res_original_id attribute."""
        return self._res_original_id
    
    @property
    def res_type(self):
        """Get res_type attribute."""
        return self._res_type
    
    @property
    def chain(self):
        """Get chain attribute."""
        return self._chain
    
    @property
    def atom_polarity(self):
        """Get atom_polarity attribute."""
        return self._atom_polarity

    def _set_radius(self):
        """Set radius.
        
        This method sets the various atomic radii defined in the Atom class.

        IMPORTANT!!!
        In the original PRO_ACT code proline nitrogen were given the same
        radii as carbon. This is not yet implemented in this code.

        """
        if self.atom_polarity == None:
            raise MacroMoleculeAtomError,"set_radius req. polarity to be set"
        radii = R.radii[self.atom_type.upper()]
        self.char_apol_radius = radii['charApolRadius']
        if self.atom_polarity == 0:
            self.char_pol_radius = radii['charPolRadius']
    
            self.char_pol_radius_max = radii['charPolRadiusMax']
        self.close_radius = radii['closeRadius']
        self.far_radius = radii['farRadius']
        self.probe_radius = radii['probeRadius']
    
    def _set_atom_polarity(self):
        """Set atom_polarity cooridinate.

        IMPORTANT!
        In the original PRO_ACT code proline nitrogen were set to apolar.
        This is not yet implemented in this code.

        """
        if (self.atom_type == 'O'
            or self.atom_type == 'N'
            or self.atom_type == 'H2O'):
            self._atom_polarity = 0  # polar
        else:
            self._atom_polarity = 2  # apolar
 
    def in_contact(self, target):
        """
        This method checks to see if the Atom is in contact with the target
        Atom. The function returns an integer based on the type of contact.
        The types of contacts are defined as:
        1) No contact     - (-1)
        2) Polar contact  - (1)
        3) Apolar contact - (2)

        target: Atom

        Atoms of the same residue/ligand are never in contact. Atoms part of
        the same peptide bond are never in contact.

        returns int
        """
        # Check to see if the atoms are adjecent (in terms of connectivity)
        # which would result in them not being in contact
        if isinstance(target, MacroMoleculeAtom):
            # both the atom and the target are macroMoleculeAtoms
            if self._molecule_instance == target._molecule_instance:
                # the atom and the target are part of the same macroMolecule
                    if self._res_original_id == target._res_original_id:
                        # same residue -> no contact
                        return -1
                    if ( self._res_original_id - target._res_original_id
                        == 1):
                        # target is in residue before self
                        if ( (self._atom_org_name == ' N  ')
                            and
                            ( target._atom_org_name == ' CA '
                            or target._atom_org_name == ' C  '
                            or target._atom_org_name == ' O  ')):
                            # atoms physically connected -> no contact
                            return -1
                        if ( (self._atom_org_name == ' CA ')
                            and
                            ( target._atom_org_name == ' C  ')):
                            # atoms physically connected -> no contact
                            return -1
                    if ( target._res_original_id - self._res_original_id
                        == 1):
                        # self is in residue before target
                        if ( (target._atom_org_name == ' N  ')
                            and
                            ( self._atom_org_name == ' CA '
                            or self._atom_org_name == ' C  '
                            or self._atom_org_name == ' O  ')):
                            # atoms physically connected -> no contact
                            return -1
                        if ( (target._atom_org_name == ' CA ')
                            and
                            ( self._atom_org_name == ' C  ')):
                            # atoms physically connected -> no contact
                            return -1
        if self._sub_atom_type == 2 and target._sub_atom_type == 2:
            # both the atom and the target are LigandAtoms
            if self._molecule_instance == target._molecule_instance:
                # the atom and the target are part of the same Ligand
                return -1
        return Atom.in_contact(self, target)

        
# Test MacroMoleculeAtom class by 'python MacroMoleculeAtom.py'
if __name__ == "__main__":
    import doctest
    doctest.testmod()
    import unittest
    suite = doctest.DocFileSuite('doc_tests/test_MacroMoleculeAtom.txt',
        optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE)
    unittest.TextTestRunner().run(suite)
