from SurfacePoint import SurfacePoint

class Surface(object):
    """Class Surface

    The Surface class is a container which holds instances of the class
    SurfacePoint in the attribute surface_point_list. The class also holds the
    attributes sa_point_count (solvent accessible point count) and
    pa_point_count (probe accessible point count), which hold information
    about the accessibility of the surface.

    Attributes:
    surface_point_list   -- list of surface points
    sa_point_count       -- number of solvent accessible surface points in
                            surface_point_list
    pa_point_count       -- number of probe accessible surface points in
                            surface_point_list
    
    Methods:
    __init__             -- Initialize an instance.

    Examples:
    >>> from Surface import Surface
    >>> from UnitSurfacePoints import UnitSurfacePoints
    >>> from SurfacePoint import SurfacePoint
    >>> from Point import Point
    >>> u = UnitSurfacePoints()
    >>> c = Point(1.0, 1.0, 1.0)   # The centre (of an atom)
    >>> s = Surface(u, c, radius=3.5, accessibility=0, flag=0)
    >>> s.sa_point_count
    126
    >>> s.pa_point_count
    0

    """
    
    def __init__(self, unit_surface_points, centre_point, radius,
                 accessibility, flag):
        self._surface_point_list = []
        for i, unit_point in enumerate(unit_surface_points.point_list):
            self._surface_point_list.append(SurfacePoint(centre_point,
                                            unit_point, i, radius, accessibility,
                                            flag))

    @property
    def surface_point_list(self):
        """Get surface_point_list attribute."""
        return self._surface_point_list
    
    @property
    def sa_point_count(self):
        """Get sa_point_count (solvent accessible point count) attribute."""
        asa = self._loopReturn(2)  # apolar solv acc surf count
        psa = self._loopReturn(0)  # polar solv acc surf count
        return asa + psa
    
    @property
    def pa_point_count(self):
        """Get pa_point_count (probe accessible point count) attribute."""
        pa = self._loopReturn(-1)  # probe acc surf count
        pa1 = self._loopReturn(-3)  # also probe acc surf count
        return pa + pa1

    def _loopReturn(self, accessibility):
        """Function loopReturn
        
        accessibility: int

        This method is used internally to loop through the surfacePoints in
        surface_point_list and return the number of points with the specified
        accessibility.
        
        returns int

        """
        count = 0
        for point in self._surface_point_list:
            if point._accessibility == accessibility:
                count = count + 1
        return count
    
    def remove_inaccessible_points(self):
        """
        Remove inaccessible surface points (accessibility == 1).
        """
        self._surface_point_list = [ x for x in self._surface_point_list if x._accessibility != 1]
    
# Test the class by running the command 'python Surface.py'
if __name__ == "__main__":
    import doctest
    doctest.testmod()
    import unittest
    suite = doctest.DocFileSuite('doc_tests/test_Surface.txt',
        optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE)
    unittest.TextTestRunner().run(suite)
