#!/usr/bin/python
from pyx import *
from math import sqrt
import os,sys
from optparse import OptionParser

class changesymbol(graph.style.symbol):

    def __init__(self, sizecolumnname="size", colorcolumnname="color",
                       gradient=color.gradient.Rainbow,
                       symbol=graph.style.symbol.circle,
                       symbolattrs=[deco.filled],#deco.stroked([color.gray.black])],
                       **kwargs):
        # add some configuration parameters and modify some other
        self.sizecolumnname = sizecolumnname
        self.colorcolumnname = colorcolumnname
        self.gradient = gradient
        graph.style.symbol.__init__(self, symbol=symbol, symbolattrs=symbolattrs, **kwargs)

    def columnnames(self, privatedata, sharedata, agraph, columnnames):
        # register the new column names
        if self.sizecolumnname not in columnnames:
            raise ValueError("column '%s' missing" % self.sizecolumnname)
        if self.colorcolumnname not in columnnames:
            raise ValueError("column '%s' missing" % self.colorcolumnname)
        return ([self.sizecolumnname, self.colorcolumnname] +
                graph.style.symbol.columnnames(self, privatedata,
                                               sharedata, agraph, columnnames))

    def drawpoint(self, privatedata, sharedata, graph, point):
        # replace the original drawpoint method by a slightly revised one
        if sharedata.vposvalid and privatedata.symbolattrs is not None:
            x_pt, y_pt = graph.vpos_pt(*sharedata.vpos)
            color = self.gradient.getcolor(point[self.colorcolumnname])
            privatedata.symbol(privatedata.symbolcanvas, x_pt, y_pt,
                               privatedata.size_pt*point[self.sizecolumnname],
                               privatedata.symbolattrs + [color])

def make_graph(xmin,xmax,ymin,ymax):
    global g
    xlen=xmax-xmin
    ylen=ymax-ymin
    if xlen >= ylen:
        ratio = ylen/float(xlen)
        xsize = imagesize; ysize = imagesize * ratio
    else: 
        ratio = xlen/float(ylen)
        xsize = imagesize * ratio; ysize = imagesize
    if multcheck == 1:
        default=1.05
    else:
        default=1
    g=graph.graphxy(width=xsize, height=ysize,
                    x=graph.axis.linear(title='Residue (Apolar contacts)', min=xmin, max=xmax+1),
                    y=graph.axis.linear(title='Residue (Polar contacts)', min=ymin, max=ymax+1),
                    key=graph.key.key(pos=None,hpos=0.5,vpos=default+(1.0/imagesize)))
#
# Bug in pyx key module! need to overwrite the default pos value to use hpos and vpos
# Additional bug/missing feature - TeX appparently can't understand '_' when writing titles

#####################################
#   Command line Options            #
#####################################

usage = "python %prog [options] input.csv [output.pdf]"
version = "%prog 1.0.1"
parser = OptionParser(usage=usage, version=version)
    
parser.add_option("-v", "--verbose",
                  action="store_true", dest="verbose", default=False,
                  help="print status messages to stdout")
parser.add_option("-d", "--debug",
                  action="store_true", dest="debug", default=False,
                  help="Include debugging options")
parser.add_option("-i", "--imagesize",
                  dest="imagesize", default=20,
                  help="image size (in cm, default=20)", type="float", 
                  metavar="SIZE")
parser.add_option('-x', '--xrange', action='store',default='-:',metavar='xrange',
                  help="First range of residues to compare contacts. For residues 20-40 of chain A, use 'A-20:40', For all of chain 1 and chain B, use '1B-:' To use everything, leave this blank (default).")
parser.add_option('-y', '--yrange', action='store',default='-:',metavar='yrange',
                  help="Second range of residues to compare contacts. For residues 20-40 of chain A, use 'A-20:40', For all of chain 1 and chain B, use '1B-:' To use everything, leave this blank (default).")
parser.add_option('-o', '--output', action='store',default='eps',metavar='output',
                  help="Output format (ps, eps or pdf)")
parser.add_option("-w", "--include_waters",
                  action="store_true", dest="water", default=False,
                  help="Include contacts between water molecules and protein residues")

(options, args) = parser.parse_args()

y = options.xrange.split('-')
chain=y[0]
if y[1].split(':')[0]:
    low = int(y[1].split(':')[0])
else: low = 0
if y[1].split(':')[1]:
    high = int(y[1].split(':')[1])
else: high = 9999
xopt= [chain, low, high]

y = options.yrange.split('-')
chain=y[0]
if y[1].split(':')[0]:
    low = int(y[1].split(':')[0])
else: 
    low = 0
if y[1].split(':')[1]:
    high = int(y[1].split(':')[1])
else: 
    high = 9999
yopt= [chain, low, high]

assert options.output.lower() in ('ps','eps','pdf')
imagesize=options.imagesize
filename=args[0]
multcheck=1
    
###############################################################
# Main file                                                   #
###############################################################
file=open(filename,'r').readlines()
Plist=[]; Alist=[]; reslist=[]
for data in file[1:]: # Determine the number of residues and the largest number of polar and apolar cavities from the contact file
    newdata = data.split(',')
    num1,res1,chain1,num2,res2,chain2,polar,apolar=newdata
    polar=int(polar); apolar=int(apolar); res1= int(num1)
    Plist.append(polar); Alist.append(apolar); reslist.append(res1)
Plist.sort(); Alist.sort(); reslist.sort()

resmin = reslist.pop(0)
resmax = reslist.pop()
pmax = Plist.pop()
amax = Alist.pop()

poslist=[]
for data in file[1:]: # extract the necessary data from the graph in a list (residue 1, residue 2, # of polar/apolar contacts, colour code)
    newdata = data.split(',')
    num1,res1,chain1,num2,res2,chain2,polar,apolar=newdata
#    print res1,chain1,res2,chain2,polar,apolar
#    print res1, res2
    if xopt[0]: # only use entries in the correct chain, if a chain is specified
        if chain1 not in xopt[0]:
            continue
    if yopt[0]:
        if chain2 not in yopt[0]:
            continue
    if not options.water: # do not use water contacts unless specified
        if res2 == '"HOH"':
            continue
    if res2 == '"PPP"': # do not use probe contacts
        continue
    num1=int(num1); num2=int(num2); polar=int(polar); apolar=int(apolar)
    if num1 < num2:
        if polar > 0:
            poslist.append([num1,num2,polar/float(pmax)*2,0])
        if apolar > 0:
            poslist.append([num2,num1,sqrt(apolar)/float(amax)*8,1])
    else:
        if polar > 0:
            poslist.append([num2,num1,polar/float(pmax)*2,0])
        if apolar > 0:
            poslist.append([num1,num2,sqrt(apolar)/float(amax)*8,1])

for item in (xopt,yopt): # set the graph size to the first and last residue, if not smaller than this
    if item[1] < resmin:
        item[1] = resmin
    if item[2] > resmax:
        item[2] = resmax

make_graph(xopt[1],xopt[2],yopt[1],yopt[2]) # open the pyx graph of the chosen size

g.plot(graph.data.points(poslist, x=1,y=2,size=3,color=4, 
                       title=r"Polar and Apolar Contacts: Largest red dot is %s polar contacts, largest blue dot is %s apolar contacts"%(pmax, amax)), [changesymbol()])
if options.output.lower() == 'pdf':
    g.writePDFfile(filename.split('.')[0])
elif options.output.lower() == 'ps':
    g.writePSfile(filename.split('.')[0])
elif options.output.lower() == 'eps':
    g.writeEPSfile(filename.split('.')[0])
print "Graph printed to %s.%s" %(filename.split('.')[0],options.output.lower())
